# ==================================================================
#                  _____ _                           
#                 / ____| |                          
#                | |    | |__  _ __ ___  _ __   ___  
#                | |    | '_ \| '__/ _ \| '_ \ / _ \ 
#                | |____| | | | | | (_) | | | | (_) |
#                 \_____|_| |_|_|  \___/|_| |_|\___/ 
#                                                    fidle.Chrono
# ==================================================================
# A simple class pour implement a stopwatch (and a bit more)
# Jean-Luc Parouty CNRS/MIAI/SIMaP 2022

import datetime
import fidle.config as config
import fidle.utils  as utils


class Chrono:

    def __init__(self):
        __version__   = config.VERSION
        self.reset_chrono()


    def start(self, id='default'):
        '''
        Start a chronometer
        args:
            id : chronometer name ('default')
        return:
            nothing
        '''
        self.start_time[id] = datetime.datetime.now()


    def stop(self, id='default'):
        '''
        Stop a chronometer
        args:
            id : chronometer name ('default')
        return:
            nothing
        '''
        self.end_time[id]   = datetime.datetime.now()


    def get_delay(self, id='default', format='str'):
        '''
        Return a delay from a chronometer
        If chronometer is stopped, return delay=(stop-start)
        If chronometer is running, return delay=(now()-start)
        args:
            id     : chronometer name                        ('default')
            format : delay format, 'str', 'seconds', 'human' (str)
        return:
            formated delay
        '''
        t1 = self.start_time[id]
        t2 = self.end_time.get(id, datetime.datetime.now())
        dt = t2 - t1
        # ---- in seconds
        if format=='seconds': 
            return round(dt.total_seconds(),2)
        # ---- str, seconds without microseconds
        if format=='str':
            dt = dt - datetime.timedelta(microseconds=dt.microseconds)
            return str(dt)
        # ---- Human
        if format=='human':
            return utils.hdelay(dt)
        # ---- timedelta
        return dt


    @staticmethod
    def now(format='%d/%m/%y %H:%M:%S'):
        '''
        Return a formated string for current date/time
        args:
            format : a strftime format ('%d/%m/%y %H:%M:%S')
        return:
            current formated date/time
        '''
        t = datetime.datetime.now()
        if format is None:  return t
        return t.strftime(format)
    

    @staticmethod
    def tag_now():
        '''
        Return a string tag for current date/time
        args:
            nothing
        return:
            current formated date/time tag as '%Y-%m-%d_%Hh%Mm%Ss'
        '''
        return datetime.datetime.now().strftime("%Y-%m-%d_%Hh%Mm%Ss")


    def get_start(self, id='default', format='%d/%m/%y %H:%M:%S'):
        '''
        Return start time from a chronometer
        args:
            id     : chronometer name         ('default')
            format : strftime format or None  ('%d/%m/%y %H:%M:%S')
        return:
            formated start time or datetime if format is None
        '''
        t1 = self.start_time[id]
        if format is None:  return t1
        return t1.strftime(format)


    def get_end(self, id='default', format='%d/%m/%y %H:%M:%S'):
        '''
        Return end time from a chronometer
        args:
            id     : chronometer name         ('default')
            format : strftime format or None  ('%d/%m/%y %H:%M:%S')
        return:
            formated start time or datetime if format is None
        '''
        t2 = self.end_time[id]
        if format is None:  return t2
        return t2.strftime(format)


    def reset_chrono(self):
        '''
        Reset all chronometers.
        Remove all start/stop entries.
        args:
            nothing
        return:
            nothing
        '''
        self.start_time = {} 
        self.end_time   = {}


    def show(self,id='default'):
        '''
        Print current duration for a chronometer in seconds
        args:
            id : chronometer name ('default')
        return
            nothing, just print
        '''
        print('Duration : ', self.get_delay(id,format='seconds'),'seconds')





# ------------------------------------------------------------------
#   ______ _     _ _                    _           _       
#  |  ____(_)   | | |          /\      | |         (_)      
#  | |__   _  __| | | ___     /  \   __| |_ __ ___  _ _ __  
#  |  __| | |/ _` | |/ _ \   / /\ \ / _` | '_ ` _ \| | '_ \ 
#  | |    | | (_| | |  __/  / ____ \ (_| | | | | | | | | | |
#  |_|    |_|\__,_|_|\___| /_/    \_\__,_|_| |_| |_|_|_| |_|
#                                                  Fidle admin (fid)
# ------------------------------------------------------------------
# A simple class to admin fidle bizness : fid command
# Jean-Luc Parouty CNRS/MIAI/SIMaP 2022


from ast import arg
from asyncio import format_helpers
from genericpath import isdir
import os
import sys
import traceback
import argparse
from typing_extensions import Required
import yaml
import importlib

from IPython.display import display
import pandas as pd


import fidle.config as config
import fidle.utils  as utils
from fidle.Chrono           import Chrono
from fidle.TocBuilder       import TocBuilder
from fidle.MagicCooker      import MagicCooker
from fidle.Installer        import Installer
from fidle.TimeSheetManager import TimeSheetManager


__version__        = config.VERSION
use_interactivity  = False                # by default

usage_end_user='''
    fid

      check                                   Check environment
            [--directory    <directory> ]        The place to check            (.)

      install                                 Install notebooks and datasets
            [--notebooks   <ress. name> ]        Notebooks ressource name      (default)
            [--datasets    <ress. name> ]        Datasets ressource name       (default)
            [--install_dir <directory>  ]        The place to install          (.)
            [--quiet                    ]        Quiet mode

      install_notebooks                       Install notebooks
            [--notebooks   <ress. name> ]        Notebooks ressource name      (default)
            [--install_dir  <directory> ]        The place to install          (.)
            [--quiet                    ]        Quiet mode

      install_datasets                        Install datasets
            [--datasets    <ress. name> ]        Datasets ressource name       (default)
            [--install_dir  <directory> ]        The place to install          (.)
            [--quiet                    ]        Quiet mode

      show_branch                             Show available notebooks and datasets branches

      help | --help                            Help             
    
    Examples :
    --------
    1/ To install Fidle notebooks and datasets in the current directory :
            # fid --install

    2/ To install a new version or reinstall notebooks in the current directory :
            # Rename or remove your existant notebooks directory, then
            # fid install_notebooks
    
    '''

usage_admin_user='''
    fid

    END USER COMMANDS :

      check                                   Check environment
            [--directory    <directory> ]        The place to check            (.)

      install                                 Install notebooks and datasets
            [--notebooks   <ress. name> ]        Notebooks ressource name      (default)
            [--datasets    <ress. name> ]        Datasets ressource name       (default)
            [--install_dir <directory>  ]        The place to install          (.)
            [--quiet                    ]        Quiet mode

      install_notebooks                       Install notebooks
            [--notebooks   <ress. name> ]        Notebooks ressource name      (default)
            [--install_dir <directory>  ]        The place to install          (.)
            [--quiet                    ]        Quiet mode

      install_datasets                        Install datasets
            [--datasets    <ress. name> ]        Datasets ressource name       (default)
            [--install_dir <directory>  ]        The place to install          (.)
            [--quiet                    ]        Quiet mode

      show_branch                             Show available notebooks and datasets branches

    ADMIN COMMANDS :

      toc                                     (re)Build readme.md and readme.ipynb
            [--about_file  <about file> ]         About file                    (./fidle/about.yml)
            [--root_dir    <root dir>   ]         root dir of the notebooks     (.)

      run_ci                                  Run a continous integration campain
            [--root_dir    <root dir>   ]        Root directory                (.)
            [--campain     <ci profile> ]        Campain profile name          (./fidle/ci/run-01.yml)
            [--filter      <filter> x n ]        Filter to apply               (.*)
            [--campain_tag <tag>        ]        Campain tag                   (None)
            [--quiet                    ]        Quiet mode                

      reset_ci                                Reset a continous integration campain
            [--root_dir    <root dir>   ]         Root directory                (.)
            [--campain     <ci profile> ]         Campain profile name          (./fidle/ci/run-01.yml)
            [--filter      <filter> x n ]         Filter to apply               (.*)
            [--campain_tag <tag>        ]         Campain tag                   (None)
            [--quiet                    ]         Quiet mode
      
      tms_check                               Just Show time sheet emails
            [--profile     <tms profile> ]       Profile name                  (~/timeSheet.yml)
            [--sequence_id <sequence_id> ]       Sequence id                   (None)
            [--magickey    <magickey>    ]       MagicKey                      (None)
            [--at          <date>        ]       Start or end date, format : 21-09-2022 11:30:00 or None for now
            [--duration    <duration>    ]       Duration in minutes,          (10)

      tms_send                                Send confirmations time sheet emails
            [--profile     <tms profile> ]       Profile name                  (~/timeSheet.yml)
            [--sequence_id <sequence_id> ]       Sequence id                   (None)
            [--magickey    <magickey>    ]       MagicKey                      (None)
            [--at          <date>        ]       Start or end date, format : 21-09-2022 11:30:00 or None for now
            [--duration    <duration>    ]       Duration in minutes,          (10)
            [--no_copy                   ]       Don't save a copy of each confirmation email sent
            [--no_flag_answered          ]       Don't set the answered flag of the emails to which a confirmation is sent

      tms_unflag                              Unflag time sheet emails, ie: remove answered and seen flags
            [--profile     <tms profile> ]       Profile name                  (~/timeSheet.yml)
            [--sequence_id <sequence_id> ]       Sequence id                   (None)
            [--magickey    <magickey>    ]       MagicKey                      (None)
            [--at          <date>        ]       Start or end date, format : 21-09-2022 11:30:00 or None for now
            [--duration    <duration>    ]       Duration in minutes,          (10)

      help                                    Help
            [--admin]                            Admin usage
    
    Examples :
    --------
    1/ To install Fidle notebooks and datasets :
            # fid --install

    2/ To install a new version or reinstall notebooks :
            # Rename or remove your existant notebooks directory, then
            # fid install_notebooks
    
    '''



# ------------------------------------------------------------------
# -- Check
# ------------------------------------------------------------------
#
def do_check(args):
    '''
    Check Fidle environment.
    '''
    installer=Installer()
    installer.check(directory = args.directory)


# ------------------------------------------------------------------
# -- Toc
# ------------------------------------------------------------------
#
def do_toc(args):
        print('Update TOC in readme : \n')
        tb       = TocBuilder()
        tb.update( about_file = args.about_file, 
                   root_dir   = args.root_dir )

# ------------------------------------------------------------------
# -- reset_ci
# ------------------------------------------------------------------
#
def do_reset_ci(args):
        print('Reset continous intégration campain : \n')
        if not are_you_sure(args.quiet) : return
        mck = MagicCooker()
        mck.reset_campain( args.campain,
                           root_dir    =  args.root_dir,
                           filters     =  args.filter, 
                           campain_tag =  args.campain_tag)

# ------------------------------------------------------------------
# -- run_ci
# ------------------------------------------------------------------
#
def do_run_ci(args):
        print('Run continous intégration campain : \n')
        if not are_you_sure(args.quiet) : return
        mck = MagicCooker()
        mck.run_campain( args.campain,
                         root_dir    =  args.root_dir,
                         filters     =  args.filter, 
                         campain_tag =  args.campain_tag)


# ------------------------------------------------------------------
# -- install
# ------------------------------------------------------------------
#
def do_install(args):

        do_install_notebooks(args)
        do_install_datasets(args)


# ------------------------------------------------------------------
# -- install notebooks
# ------------------------------------------------------------------
#
def do_install_notebooks(args):
        print(f'Install Fidle notebooks in {args.install_dir} : \n')
        if not are_you_sure(args.quiet) : return
        installer=Installer()
        installer.install_ressource( catalog_name     = 'notebooks', 
                                     ressource_name   = args.notebooks, 
                                     installation_dir = args.install_dir )

# ------------------------------------------------------------------
# -- install datasets
# ------------------------------------------------------------------
#
def do_install_datasets(args):
        print(f'Install Fidle datasets in {args.install_dir} : \n')
        if not are_you_sure(args.quiet) : return
        installer=Installer()
        installer.install_ressource( catalog_name     = 'datasets', 
                                     ressource_name   = args.datasets, 
                                     installation_dir = args.install_dir )

# ------------------------------------------------------------------
# -- show_branch
# ------------------------------------------------------------------
#
def do_show_branch(args):
        installer=Installer()
        installer.show_branch('notebooks')
        installer.show_branch('datasets')



# ------------------------------------------------------------------
# -- tms show
# ------------------------------------------------------------------
#
# tms_show                                Just Show time sheet emails
#     [--profile     <tms profile> ]       Profile name                  (~/timeSheet.yml)
#     [--sequence_id <sequence_id> ]       Sequence id                   (None)
#     [--magickey    <magickey>    ]       MagicKey                      (None)
#     [--at          <date>        ]       Start or end date,   format : 21-09-2022 11:30:00
#     [--duration    <duration>    ]       Duration in minutes,          (10)

def do_tms_check(args):
    print(f'Check presences')
    tms  = TimeSheetManager(args.profile, debug=0)

    presences = tms.get_presences( sequence_id = args.sequence_id, 
                                    magickey    = args.magickey,
                                    at          = args.at, 
                                    duration    = int(args.duration) ) 

    print('Complete list of presence mails : \n')

    df = pd.DataFrame.from_dict(presences)
    df = df[ ['from','send_dt','transit','attestation_ok','sequence_id'] ]
    with pd.option_context('display.max_rows', None, 'display.max_columns', None, 'display.width', 2000):
        display(df)
    print()

    tms.save_presences(presences, save_as=args.save, default='Check')

        

# ------------------------------------------------------------------
# -- tms send
# ------------------------------------------------------------------
#
#   tms_send                                Send confirmations time sheet emails
#         [--profile     <tms profile> ]       Profile name                  (~/timeSheet.yml)
#         [--sequence_id <sequence_id> ]       Sequence id                   (None)
#         [--magickey    <magickey>    ]       MagicKey                      (None)
#         [--at          <date>        ]       Start or end date,   format : 21-09-2022 11:30:00
#         [--duration    <duration>    ]       Duration in minutes,          (10)
#         [--no_copy                   ]       Don't save a copy of each confirmation email sent
#         [--no_flag_answered          ]       Don't set the answered flag of the emails to which a confirmation is sent

def do_tms_send(args):
    print(f'Send confirmation presence emails')
    tms  = TimeSheetManager(args.profile, debug=0)

    presences = tms.get_presences(    sequence_id   = args.sequence_id, 
                                        magickey      = args.magickey,
                                        at            = args.at, 
                                        duration      = int(args.duration) ) 

    tms.send_confirmations(     presences, 
                                save_copy     = args.save_copy, 
                                flag_answered = args.flag_answered )

    tms.save_presences(presences, save_as=args.save, default='Send')

        
# ------------------------------------------------------------------
# -- tms unflag
# ------------------------------------------------------------------
#
#   tms_unflag                              Unflag time sheet emails, ie: remove answered and read flags
#         [--profile     <tms profile> ]       Profile name                  (~/timeSheet.yml)
#         [--sequence_id <sequence_id> ]       Sequence id                   (None)
#         [--magickey    <magickey>    ]       MagicKey                      (None)
#         [--at          <date>        ]       Start or end date, format : 21-09-2022 11:30:00 or None for now
#         [--duration    <duration>    ]       Duration in minutes,          (10)

def do_tms_unflag(args):
    print(f'Remove presence emails flags (answered)')

    tms  = TimeSheetManager(args.profile, debug=0)

    presences = tms.get_presences(    sequence_id   = args.sequence_id, 
                                        magickey      = args.magickey,
                                        at            = args.at, 
                                        duration      = int(args.duration) ) 

    tms.unflag( presences )

    tms.save_presences(presences, save_as=args.save, default='Unflag')



# ------------------------------------------------------------------
# -- Are you sure ?
# ------------------------------------------------------------------
#
def are_you_sure(quiet):
        if quiet :                return True
        if not use_interactivity: return True
        print('\nAre you sure ? (Y/N) ', end='')
        rep=input()
        print('')
        return (rep=='Y')


# ------------------------------------------------------------------
# Run / Module method
# ------------------------------------------------------------------
#
# Run a command, directly from python : fidle.admin.run( )
# Or via an entry point / command line
#
def run(*arguments):
    print('\n==========================================================')
    print(f'fid - Your favorite Fidle admin command :-)    (v{config.VERSION})')
    print('==========================================================\n')

    # ---- Parser
    #
    parser = argparse.ArgumentParser( description="Fidle admin command (fid)",
                                      usage=usage_end_user)

    parser.add_argument('cmd',               help=argparse.SUPPRESS,    default='help')
                 
    parser.add_argument('--directory',             dest='directory',          help=argparse.SUPPRESS,    default='.')
    parser.add_argument('--notebooks',             dest='notebooks',          help=argparse.SUPPRESS,    default='default')
    parser.add_argument('--datasets',              dest='datasets',           help=argparse.SUPPRESS,    default='default')
    parser.add_argument('--install_dir',           dest='install_dir',        help=argparse.SUPPRESS,    default='.')
    parser.add_argument('--root_dir',              dest='root_dir',           help=argparse.SUPPRESS,    default='.')
    parser.add_argument('--about_file',            dest='about_file',         help=argparse.SUPPRESS,    default='./fidle/about.yml')
    parser.add_argument('--campain',               dest='campain',            help=argparse.SUPPRESS,    default='./fidle/ci/run-01.yml')
    parser.add_argument('--filter',                dest='filter',             help=argparse.SUPPRESS,    default=['.*'], nargs='+')
    parser.add_argument('--campain_tag',           dest='campain_tag',        help=argparse.SUPPRESS,    default=None)
    parser.add_argument('--profile',               dest='profile',            help=argparse.SUPPRESS,    default='~/timeSheet.yml')
    parser.add_argument('--sequence_id',           dest='sequence_id',        help=argparse.SUPPRESS,    default=None)
    parser.add_argument('--magickey',              dest='magickey',           help=argparse.SUPPRESS,    default=None)
    parser.add_argument('--at',                    dest='at',                 help=argparse.SUPPRESS,    default=None)
    parser.add_argument('--duration',              dest='duration',           help=argparse.SUPPRESS,    default=10, type=int)
    parser.add_argument('--no_copy',               dest='save_copy',          help=argparse.SUPPRESS,    action='store_false')
    parser.add_argument('--no_flag_answered',      dest='flag_answered',      help=argparse.SUPPRESS,    action='store_false')
    parser.add_argument('--save',                  dest='save',               help=argparse.SUPPRESS,    default=None, const='default', nargs='?' )
    parser.add_argument('--quiet',                 dest='quiet',              help=argparse.SUPPRESS,    action='store_true')
    parser.add_argument('--admin',                 dest='admin',              help=argparse.SUPPRESS,    action='store_true')
    
    try:
        args  = parser.parse_args(arguments)
    except:
        print('\n')
        return

    # ---- Command
    #
    try:
        if   args.cmd=='check'                :  do_check(args)
        elif args.cmd=='install'              :  do_install(args)
        elif args.cmd=='install_notebooks'    :  do_install_notebooks(args)
        elif args.cmd=='install_datasets'     :  do_install_datasets(args)
        elif args.cmd=='show_branch'          :  do_show_branch(args)
        elif args.cmd=='toc'                  :  do_toc(args)
        elif args.cmd=='reset_ci'             :  do_reset_ci(args)
        elif args.cmd=='run_ci'               :  do_run_ci(args)
        elif args.cmd=='tms_check'            :  do_tms_check(args)
        elif args.cmd=='tms_send'             :  do_tms_send(args)
        elif args.cmd=='tms_unflag'           :  do_tms_unflag(args)
        else: 
            # Usage
            if args.admin : print(usage_admin_user)    
            else:           print(usage_end_user)    
    except KeyboardInterrupt:
        print('\nAbort !\n')
        return
    except Exception as e:
        print('\nOups...  something went wrong !')
        print(e)
        print('\nMore details :\n')
        traceback.print_exc()

        return


# ------------------------------------------------------------------
# Run / Pypi entry point
# ------------------------------------------------------------------
# fidle.admin --help
#
def main():
    global use_interactivity
    
    # ---- We are in command line mode
    use_interactivity = sys.__stdin__.isatty()

    run(*sys.argv[1:])    


# ------------------------------------------------------------------
# Run / Command line
# ------------------------------------------------------------------
# python -m fidle.admin --help
#
if __name__ == '__main__':

    # ---- We are in command line mode
    use_interactivity = sys.__stdin__.isatty()

    run(*sys.argv[1:])    




# ------------------------------------------------------------------
#   __  __             _       _____            _             
#  |  \/  |           (_)     / ____|          | |            
#  | \  / | __ _  __ _ _  ___| |     ___   ___ | | _____ _ __ 
#  | |\/| |/ _` |/ _` | |/ __| |    / _ \ / _ \| |/ / _ \ '__|
#  | |  | | (_| | (_| | | (__| |___| (_) | (_) |   <  __/ |   
#  |_|  |_|\__,_|\__, |_|\___|\_____\___/ \___/|_|\_\___|_|   
#                 __/ |                                       
#                |___/                             fidle.MagicCooker
# ------------------------------------------------------------------
# A simple class to run a set of notebooks with parameters overriding
# Jean-Luc Parouty CNRS/MIAI/SIMaP 2022

import sys,os,platform
import re
import yaml
import base64

import nbformat
from nbconvert               import HTMLExporter
from nbconvert.preprocessors import ExecutePreprocessor, CellExecutionError

import fidle.config       as     config
from fidle.Chrono         import Chrono
from fidle.MagicFeedback  import MagicFeedback
from fidle.MagicReport    import MagicReport
from fidle.TaskManager    import TaskManager


class MagicCooker:

    __version__  = config.VERSION


    def __init__(self):
        pass





    # --------------------------------------------------------------
    # Get profile
    # --------------------------------------------------------------
    #
    def __get_profile( self, 
                        profile_name = None, 
                        root_dir     = None,
                        filters      = ['.*'], 
                        campain_tag  = None):

        '''Load yaml profile'''

        # ---- Read profile
        #
        with open(profile_name,'r') as fp:
            profile=yaml.load(fp, Loader=yaml.FullLoader)

        # ---- Retreive campain, notebooks
        #
        campain   = profile['campain']
        notebooks = profile
        del notebooks['campain'] 

        # ---- Campain_dir
        #
        campain_dir = campain['directory']

        # Add tag if needed
        if campain_tag is not None: campain_dir += campain_tag

        # Absolute path from root_dir
        campain_dir = os.path.abspath( f'{root_dir}/{campain_dir}' )

        # ---- Complete campain
        #
        campain['host']         = platform.uname()[1]
        campain['location']     = os.getcwd()
        campain['profile']      = profile_name
        campain['date']         = Chrono.now(format='%A %-d %B %Y %H:%M:%S')
        campain['filters']      = filters
        campain['campain_dir']  = campain_dir

        # ---- Sub directories
        #
        campain['campain_html_dir']     = f'{campain_dir}/html'
        campain['campain_ipynb_dir']    = f'{campain_dir}/ipynb'
        campain['campain_feedback_dir'] = f'{campain_dir}/feedback'
        campain['campain_trace_dir']    = f'{campain_dir}/trace'

        # ---- Task and lock files
        #
        campain['campain_task_file']    = f'{campain_dir}/trace/task_list.json'
        campain['campain_lock_file']    = f'{campain_dir}/trace/task_list.lock'

        # What to do if a notebook already exists (default is skip)
        #
        if 'existing_notebook' not in campain : campain['existing_notebook']='skip'

        return campain, notebooks







    def reset_campain(self, 
                      profile_name = None, 
                      root_dir     = None,
                      filters      = ['.*'],
                      campain_tag  = None):
        '''
        Reset a campain - ie: remove posibles trace and task files
        args:
            profile_name : A profile in yml
            root_dir     : The root directory to be for this profile
            campain_tag  : tag to add to this campain
        return:
            None
        '''

        # ---- Expand dir
        #
        profile_name = os.path.expanduser(profile_name)
        root_dir     = os.path.expanduser(root_dir)
 
        # ---- Welcome
        #
        print(f'\n=== Fidle/MagicCooker - Reset campain')
        print(f'=== Version : {self.__version__}')

        # ---- Get profile
        #
        try:
            campain, notebooks = self.__get_profile(profile_name, root_dir, filters, campain_tag)
        except:
            print('\n*** OUPS : The profile cannot be found or retrieved... \n')
            return


        print('\nProfile           : ', campain['profile'])
        print('Campain directory : ', campain['campain_dir'],'\n')

        # ---- Remove task and lock files
        #
        try:
            os.remove(campain['campain_task_file'])
            print('Task file is removed.')
        except OSError:
            print('Task file not found.')
        try:
            os.remove(campain['campain_lock_file'])
            print('Lock file is removed.')
        except OSError:
            print('Lock file not found.')
        print('Done.\n')






    # --------------------------------------------------------------
    # Run campain
    # --------------------------------------------------------------
    #    
    def run_campain( self, 
                     profile_name = None, 
                     root_dir     = None,
                     filters      = ['.*'], 
                     campain_tag  = None):
        '''
        
        '''

        profile_name = os.path.expanduser(profile_name)
        root_dir     = os.path.expanduser(root_dir)
 
        print(f'\n=== Fidle/MagicCooker - Run campain')
        print(f'=== Version : {self.__version__}')

        # ---- Here and now ----------------------------------------
        #
        chrono = Chrono()
        chrono.start()

        # ---- About filters ---------------------------------------
        #       
        # Filters is a string
        if isinstance(filters, str) : filters = [s.strip() for s in filters.split(',')]
        # Not a list...
        if not isinstance(filters, list) :
            self.show(f"*** Oups, filters is invalid...")
            sys.exit(1)

        # ---- Retrieve profile ------------------------------------
        #
        try:
            campain, notebooks = self.__get_profile(profile_name, root_dir, filters, campain_tag)
        except:
            print('\n*** OUPS : The profile cannot be found or retrieved... \n')
            return


        # # ---- About this campain ----------------------------------
        # #
        print('\nCampain loaded :')
        print('    Date                : ', campain['date'])
        print('    Host                : ', campain['host'])
        print('    Current location    : ', campain['location'])
        print('    Profile             : ', campain['profile'])
        print('    Description         : ', campain['description'])
        print('    Version             : ', campain['version'])
        print('    Filters             : ', campain['filters'])
        print('    Campain directory   : ', campain['campain_dir'])
        print('    Existing notebook   : ', campain['existing_notebook'])
        print('    Notebook timeout    : ', campain['timeout'])

        # ---- Create campain directories
        #
        os.makedirs( campain['campain_html_dir'    ], mode=0o750, exist_ok=True)
        os.makedirs( campain['campain_ipynb_dir'   ], mode=0o750, exist_ok=True)
        os.makedirs( campain['campain_trace_dir'   ], mode=0o750, exist_ok=True)
        os.makedirs( campain['campain_feedback_dir'], mode=0o750, exist_ok=True)

        # ---- Environment vars ------------------------------------
        #
        environment_vars = campain.get('environment_vars',{})

        # ---- Create a nice task list -----------------------------
        # ----------------------------------------------------------
        #
        task_list=[]
        for task_id,notebook in notebooks.items():

            task = dict( id     = task_id,
                         after  = notebook.get('after',None),
                         status = 'desactivated' )

            if self.__match_filters(task_id, filters): task['status']='ready'
            task_list.append(task)

        # ---- Create a nice TaskManager ---------------------------
        # ----------------------------------------------------------
        #
        task_file = campain['campain_task_file' ]
        lock_file = campain['campain_lock_file' ]
       
        # ---- Create a TaskManager
        #
        tm = TaskManager( task_list, 
                          taskfile = task_file,
                          lockfile = lock_file,
                          verbose  = 0)

        # ---- Update dependencies
        #
        updated = tm.update_dependencies()
        if len(updated)>0:
            print('\nNote : For reasons of dependencies, the following tasks have been added : ')
            for task_id in updated:
                notebook = notebooks[task_id]
                path     = notebook['notebook']
                print(f'    {task_id:18s} : {path}')

        # ---- Prepare report
        #
        reporter = MagicReport(campain, notebooks, task_file)

        # ---- Run all tasks -----------------------------------------------
        # ------------------------------------------------------------------
        #
        for task_id in tm.runnable():

            notebook = notebooks[task_id]

            # ---- Get notebook infos ------------------------------
            #
            notebook_path = notebook['notebook']
            notebook_path = os.path.abspath( f'{root_dir}/{notebook_path}' )  # abs path of notebook from root_dir
            done_tag      = f'=={task_id}.done=='
            overrides     = notebook.get('overrides',{})
                
            # ---- Run notebook ------------------------------------
            # 
            feedback = self.run_notebook(  task_id            = task_id,
                                           notebook_path      = notebook_path,
                                           environment_vars   = environment_vars,
                                           overrides          = overrides,
                                           output_ipynb_dir   = campain['campain_ipynb_dir' ],
                                           output_html_dir    = campain['campain_html_dir'  ],
                                           output_tag         = done_tag,
                                           existing_notebook  = campain['existing_notebook'],
                                           timeout            = campain['timeout']
            )
            # ---- Save feedback -----------------------------------
            #
            if not feedback['status'] == 'skip': 
                # ---- Report filename
                feedback_dir  = campain['campain_feedback_dir']
                feedback_file = f'{feedback_dir}/{task_id}.json'
                # ---- Save feedback, with many infos
                mgfeed = MagicFeedback()
                mgfeed.add( task_id  = task_id,
                            campain  = campain, 
                            notebook = notebook,
                            feedback = feedback)
                mgfeed.save(feedback_file)
                print(f'    - Save feedback : {task_id}.json')

            # ---- Report ------------------------------------------
            #
            reporter.build_report()
            print(f'    - Report updated.')

        # ---- Update report
        #      Needed because task_file is saved out of step
        #
        reporter.build_report()
        
        # ---- End
        #
        print('\nDone.\n')
        tm.about_this_runner()





    # --------------------------------------------------------------
    # Run notebook
    # --------------------------------------------------------------
    #
    def run_notebook( self,
                      task_id            = None,
                      notebook_path      = None,
                      environment_vars   = {},
                      overrides          = {},
                      output_ipynb_dir   = './',
                      output_html_dir    = './',
                      output_tag         = '==done==',
                      existing_notebook  = 'remove',
                      timeout            = 6000 ):

        # ---- reporting -------------------------------------------
        #
        feedback = dict( task_id        = task_id, 
                         notebook_path = notebook_path)

        print('\nRun :', task_id)

        # ---- Notebook exist ? ------------------------------------
        #
        if task_id is None:
            print('    *** Abort : task_id cannot be None')
            feedback['status'] = 'error'
            feedback['error']  = 'Invalid task_id'
            return feedback

        if not os.path.isfile(notebook_path):
            print('    *** Abort : Notebook cannot be found : ', notebook_path )
            feedback['status'] = 'error'
            feedback['error']  = 'Notebook not found'
            return feedback

        # ---- Outputs ---------------------------------------------
        #
        notebook_dir  = os.path.dirname(  notebook_path )
        notebook_name = os.path.basename( notebook_path )
        notebook_noex = os.path.splitext( notebook_name)[0]

        output_ipynb_dir    = os.path.abspath(output_ipynb_dir)
        output_html_dir     = os.path.abspath(output_html_dir)

        output_ipynb_file   = f'{output_ipynb_dir}/{notebook_noex}{output_tag}.ipynb'
        output_html_file    = f'{output_html_dir}/{notebook_noex}{output_tag}.html'

        if (os.path.isfile(output_ipynb_file) or os.path.isfile(output_html_file) ) :
            # Output exist...
            if existing_notebook=='skip':
                # Skip : abort run
                print('    - Notebook output exist : Skip')
                feedback['status'] = 'skip'
                return feedback
            else:
                # Remove it
                print('    - Output exist : Remove it')
                if os.path.isfile(output_ipynb_file) : os.remove(output_ipynb_file)
                if os.path.isfile(output_html_file)  : os.remove(output_html_file)

        # ---- Set environment vars --------------------------------
        #
        to_unset=[]
        if isinstance(environment_vars, dict):
            print('    - Environment vars:')
            for name,value in environment_vars.items():
                # Set env
                os.environ[name] = str(value)
                # For cleaning
                to_unset.append(name)
                # About
                print(f'        {name:30s} = {value}')

        # ---- Set Overrides ---------------------------------------
        #
        print('    - Overrides :')
        if not isinstance(overrides, dict): overrides = {}

        # Probably add an override for task_id
        if not 'run_id' in overrides: overrides['run_id'] = task_id

        for name,value in overrides.items():
            # Default : no nothing
            if value=='default' : continue
            #  Set env
            env_name  = f'FIDLE_OVERRIDE_{name}'
            env_value = str(value)
            os.environ[env_name] = env_value
            # For cleaning
            to_unset.append(env_name)
            # About
            print(f'        {env_name:30s} = {env_value}')
                
        # ---- Reporting -------------------------------------------
        #
        feedback['environment_vars']  = environment_vars
        feedback['overrides']         = overrides

        # ---- Ok, Ready ? -----------------------------------------
        #
        chrono = Chrono()
        chrono.start()

        # ---- Run it ! --------------------------------------------
        #
        print('    - Notebook start...')
                
        try:
            notebook = nbformat.read( notebook_path, nbformat.NO_CONVERT)
            ep = ExecutePreprocessor(timeout=timeout, kernel_name="python3")
            ep.preprocess(notebook,  {'metadata': {'path': notebook_dir}})
            feedback['status'] = 'ok'
        except CellExecutionError as e:
            output_tag = f'=={task_id}.ERROR=='
            print('*'*60)
            print('** Oups, an error occured : ',type(e).__name__)
            print('*'*60)
            feedback['status'] = 'error'
            feedback['error']  = 'Notebook error'

        except Exception as e:
            notebook  = None
            print('*'*60)
            print('** AAARG.. A fatal error occured : ',type(e).__name__)
            print('*'*60)
            feedback['status'] = 'error'
            feedback['error']  = 'Fatal error'

        print('    - Notebook done.')
        
        # ---- Top chrono - Stop
        #
        chrono.stop()
        print(f'    - Duration    : {chrono.get_delay()}' )
        feedback['duration']  = chrono.get_delay(format='seconds')
        feedback['start']     = chrono.get_start()
        feedback['end']       = chrono.get_end()


        # ---- Clean env and overrides -----------------------------
        #
        for env_name in to_unset:
            del os.environ[env_name]

        # ---- Do nothing more if notebook died -------------------------
        #
        if notebook is None: 
            return feedback

        # ---- Check for images to embed ---------------------------
        #      We just try to embed <img src="..."> tag in some markdown cells
        #      Very fast and suffisant for header/ender.
        #
        for cell in notebook.cells:
            if cell['cell_type'] == 'markdown':
                cell.source = self.__images_embedder(cell.source, notebook_dir)

        # ---- Save notebook as ipynb -----------------------------------------
        #
        if output_ipynb_dir is not None:
            # Notebook filename (rebuilt because tag can be changed)
            output_ipynb_file = f'{output_ipynb_dir}/{notebook_noex}{output_tag}.ipynb'
            os.makedirs(output_ipynb_dir, mode=0o750, exist_ok=True)
            # Save
            with open(output_ipynb_file, mode="w", encoding='utf-8') as fp:
                print(f'    - Save ipynb  : {notebook_noex}{output_tag}.ipynb')
                nbformat.write(notebook, fp)

        # ---- Save notebook as html ------------------------------------------
        #
        if output_html_dir is not None:

            # ---- Convert notebook to html
            exporter = HTMLExporter()
            exporter.template_name = 'classic'
            (body_html, resources_html) = exporter.from_notebook_node(notebook)
            
            # ---- Embed images
            # Note : Pas du tout efficace.. et a priori inutile car le notebook a dejà été traité
            # body_html = self.__images_embedder(body_html)
            # print('    - html : images embedded.')

            # ---- Save
            # HTML filename (rebuilt because tag can be changed)
            output_html_file = f'{output_html_dir}/{notebook_noex}{output_tag}.html'
            os.makedirs(output_html_dir, mode=0o750, exist_ok=True)
            # Save
            with open(output_html_file, mode='wb') as fp:
                print(f'    - Save html   : {notebook_noex}{output_tag}.html')
                fp.write(body_html.encode("utf-8"))

        # ---- Reporting
        #
        feedback['output_ipynb_file'] = f'{notebook_noex}{output_tag}.ipynb'
        feedback['output_html_file']  = f'{notebook_noex}{output_tag}.html'

        # ---- Back to the old world..
        #
        return feedback


    # --------------------------------------------------------------
    # -- Match Filter
    # --------------------------------------------------------------
    #
    def __match_filters(self, task_id, filters):
        # We will look for a good reason to accept

        # Filters is None : Ok
        if filters is None : return True

        # For each filter
        for filter in filters:
            if re.match(filter, task_id): return True

        # Arrived here, we found no reason to accept
        return False
    
    

    def __get_base64_image(self, filename):
        '''
        Read an image file and return as a base64 encoded version
        params:
            filename : image filemane
        return:
            base 64 encoded image
        '''
        with open(filename, "rb") as image_file:
            img64 = base64.b64encode(image_file.read())
        src="data:image/svg+xml;base64,"+img64.decode("utf-8") 
        return src



        
    def __images_embedder(self, html, notebook_dir):
        '''
        Images embedder. Search images src="..." link and replace them
        by base64 embedded images.
        params:
            html: input html
        return:
            output html
        '''
        for img_tag in re.findall('.*(<img .*>).*', html):
            for src_tag in re.findall('.*src=[\'\"](.*)[\'\"].*', img_tag):
                # ---- Is is a data: tag ?
                if src_tag.startswith('data:'): continue
                # ---- Change file image to data:
                src_path = f'{notebook_dir}/{src_tag}'
                src_b64 = self.__get_base64_image(src_path)
                # ---- Replace it
                img_b64 = img_tag.replace(src_tag, src_b64)
                html = html.replace(img_tag,img_b64)
        return html





# ------------------------------------------------------------------
#   _______        _    __  __                                   
#  |__   __|      | |  |  \/  |                                  
#     | | __ _ ___| | _| \  / | __ _ _ __   __ _  __ _  ___ _ __ 
#     | |/ _` / __| |/ / |\/| |/ _` | '_ \ / _` |/ _` |/ _ \ '__|
#     | | (_| \__ \   <| |  | | (_| | | | | (_| | (_| |  __/ |   
#     |_|\__,_|___/_|\_\_|  |_|\__,_|_| |_|\__,_|\__, |\___|_|   
#                                                 __/ |          
#                                                |___/   
#                                                  fidle.TaskManager                          
# ------------------------------------------------------------------
#     
# A simple class to manage a list of tasks in a distributed way.
# Jean-Luc Parouty CNRS/MIAI/SIMaP 2022

# Task list is an array of dict
#
# <task list> = [ <task> ]
# <task>      = {  id     = <task id>,
#                  after  = [ <task id regex>, ...]
#                  status = "ready" | "running" | "done" | "desactivated",
#                  runner = <runner id>, 
#                  start  = <date time>,
#                  end    = <daate time>
#  }



import os,platform,sys
import json
import datetime, time
import re
import math
import random
from random  import uniform

import fidle.config
from fidle.Chrono import Chrono


class LockException(Exception):
    pass


class TaskManager:

    __version__  = fidle.config.VERSION


    # --------------------------------------------------------------
    # -- Init TaskManager
    # --------------------------------------------------------------
    #
    def __init__(self, tasklist         = [], 
                       taskfile         = './taskfile.json', 
                       lockfile         = './taskfile.lock', 
                       verbose          = 0,
                       sleep_start      = 4,
                       lock_check_delay = 0.100,
                       lock_retry_min   = 0.8,
                       lock_retry_max   = 1.2,
                       lock_retry_k     = 4 ):
        self.verbose = verbose
        self.whoami           = platform.uname()[1]+'-'+str(os.getpid())
        self.task_list        = tasklist
        self.task_file        = taskfile
        self.lock_file        = lockfile
        self.sleep_start      = sleep_start
        self.lock_check_delay = lock_check_delay
        self.lock_retry_min   = lock_retry_min
        self.lock_retry_max   = lock_retry_max
        self.lock_retry_k     = lock_retry_k

        # For monitoring (about)
        self.duration         = 0
        self.max_retry        = 0
        self.done_tasks       = 0

        # For random
        random.seed(os.getpid())

        # Locked state
        self.locked       = False

        # Checks afters
        self.__check_afters()

        # Verbosity
        if self.verbose>0 : self.show('Init TaskManager, done.')



    def __check_afters(self):
        for t in self.task_list:
            # No after    : add after None
            t['after'] = t.get('after', None)
            # After empty : set None
            if t['after'] == '' : t['after']=None
            # After None  : ok
            if t['after'] is None : continue
            # After is a string
            if isinstance(t['after'], str) : t['after'] = [s.strip() for s in t['after'].split(',')]
            # Not a list...
            if not isinstance(t['after'], list) :
                self.show(f"*** Oups, after directive for task [{t['id']}] is invalid...")
                sys.exit(1)



    # --------------------------------------------------------------
    # -- Verbosity
    # --------------------------------------------------------------
    #

    def show(self, *what):
        '''
        Print informations with a <time> <host-pid> prefix
        if an info item is "tasks", then print tasks list. 
        Params:
            *whats : list of informations
        Return:
            None
        '''
        when = datetime.datetime.now().strftime('%H:%M:%S.%f')
        for w in what:
            if w=='tasks':
                print(f'\n{when} {self.whoami} : Tasks list :')
                print( json.dumps(self.task_list, ensure_ascii=False, indent=4))
            else:
                print(f'{when} {self.whoami} : {w}')
            
    def about_this_runner(self):
        '''
        Display a summary of the runnable() session
        Params:
            None
        Returns:
            None
        '''
        self.show('About this runner :')
        self.show(f'    Done tasks     : {self.done_tasks}' )
        self.show(f'    Lock retry     : {self.max_retry}'   )
        self.show(f'    Total duration : {self.total_duration}'    )


    # --------------------------------------------------------------
    # -- Runnable
    # --------------------------------------------------------------
    #
    def runnable(self):
        '''
        Task generator, allowing to iterate on the task list.
        Each time an item is given, it is marked 'runnning' while 
        the previous one is marked 'done'.
        Params:
            None
        Returns:
            Sequence of task_id 
        '''
        # Few monitoring
        chrono = Chrono()
        chrono.start('total')
        self.max_retry      = 0
        self.done_tasks     = 0
        self.total_duration = 0

        # In order not to start all at the same time
        time.sleep(uniform(0, self.sleep_start))

        # Let's go...
        task_id  = ''
        duration = 0
        now      = chrono.now()
        while task_id is not None:

            # ---- Retrieve tasks list
            #
            self.__lock(retry=15)
            self.__load_tasks()

            # ---- Set current task as : done
            #
            self.__set_done(task_id, now, duration)
            
            # ---- Pick a new task
            #
            task_id  = self.__pick_task()

            # ---- Set new task as : running 
            #
            self.__set_running(task_id, now)

            # ---- Save tasks list
            #
            self.__save_tasks()
            self.__unlock()

            # ---- Yield new task
            #
            if task_id is not None :
                if self.verbose>0 :    self.show(f'Task [{task_id}] is starting')
                chrono.start('task')
                yield task_id
                duration = chrono.get_delay('task', format='seconds')
                now      = chrono.now()
                if self.verbose>0 :    self.show(f'Task [{task_id}] is done after {duration}s')
        
        self.total_duration = chrono.get_delay('total', format='seconds')



    # --------------------------------------------------------------
    # -- Lock
    # --------------------------------------------------------------
    #
    def __lock(self, retry=15):
        '''
        Try to take the hand, by creating a lock file.
        Wait and retry if a lock file still exist.
        Params:
            retry : Number of tries (5)
        Return:
            True if lock is created, else False.
        '''
        if self.verbose>1 : self.show("TaskManager - Lock")

        # ---- Still locked
        #
        if self.locked:
            if self.verbose>1 : self.show("    Still locked.. Etonnant, non ?")
            return self.locked

        # ---- No locked, we try to
        #
        n_retry=0
        while (self.locked==False) and (n_retry<retry):

            # ---- Is there a lock file ?
            #
            if not os.path.isfile(self.lock_file): 
                # It's free, so we can create a lockfile
                open(self.lock_file, "w").write(self.whoami)
                if self.verbose>1 : self.show('    No lock file : we try to put one')
                # We wait a moment and we check that we have the hand
                # Maybe someone else was faster!
                if self.verbose>1 : self.show('    Wait and check it')
                time.sleep(self.lock_check_delay)
                try: 
                    locker_id   = open(self.lock_file).readline()
                    self.locked = (locker_id == self.whoami)
                except:
                    if self.verbose>1 : self.show('    We got passed !')
                    self.locked = False

            if not self.locked:
                # The place is not free, we wait a moment and try again
                t = uniform(self.lock_retry_min, self.lock_retry_max) * math.e**(n_retry/self.lock_retry_k)
                if self.verbose>1 : self.show(f'    Lock file exist, wait {t:0.2f}s and retry ({n_retry})...')
                time.sleep(t)
                n_retry += 1
                # for informations
                self.max_retry = max(self.max_retry, n_retry)

        if self.verbose>1 : self.show(f'    Locking is: {self.locked}')
        if not self.locked: 
            raise LockException('Lock failed... (increase the number of retries ?)')
        else:
            if self.verbose>1 : self.show(f'    Nb retry : {n_retry}')
            return self.locked


    # --------------------------------------------------------------
    # -- Unlock
    # --------------------------------------------------------------
    #
    def __unlock(self):
        if self.verbose>1 : self.show('TaskManager - Unlock')
        os.unlink(self.lock_file)
        self.locked = False
        if self.verbose>1 : self.show(f'    Lock file removed ({self.lock_file})')


    # --------------------------------------------------------------
    # -- Load tasks
    # --------------------------------------------------------------
    #
    def __load_tasks(self):
        '''
        Load tasks file. Create the file if it does not exist.
        '''
        if self.verbose>1 : self.show(f'TaskManager - Load tasks')

        # ---- Task file exist : Read and return it
        #
        if os.path.isfile(self.task_file):
            with open(self.task_file) as fp:
                self.task_list = json.load(fp)
            if self.verbose>1 : self.show(f'    Loaded ({self.task_file})')
            return self.task_list

        # ---- Doesn't exist : Save it
        #
        if self.verbose>1 : self.show('    Tasks file does not exist, create it...')
        self.__save_tasks()


    # --------------------------------------------------------------
    # -- Pick task
    # --------------------------------------------------------------
    #
    def __pick_task(self):
        '''
        Pick a task from the list, checking the dependency constraints.
        Task is picked from self.task_list.
        params:
            none
        return:
            picked task_id
        '''
        if self.verbose>1 : self.show('TaskManager - Pick a task')

        # ---- Search a task to run...
        #
        picked_task = None
        for task in self.task_list:
            # Check if status is "ready" 
            if task['status'] != 'ready': continue 
            # Check for dependencies (after)
            if self.__check_dependency(task) is False : continue
            # Ok, this one is runnable
            picked_task = task
            break
        
        # ---- No tasks are available...
        #
        if picked_task is None : 
            if self.verbose>1 : self.show('    Cannot find a new task')
            return None

        # ---- return picked task id
        #
        picked_task_id = picked_task['id']
        if self.verbose>1 : self.show(f"    Selected task is : {picked_task_id}")
        return picked_task_id


    # --------------------------------------------------------------
    # -- Get task
    # --------------------------------------------------------------
    #
    def __get_task(self, task_id):
        for task in self.task_list:
            if task['id']==task_id: return task
        return None

    # --------------------------------------------------------------
    # -- Set running
    # --------------------------------------------------------------
    #
    def __set_running(self, task_id, now):
        '''
        Set given task_id as "running"
        Complete status, runner and start attributes.
        Params:
            task_id : Task id
        Return:
            None
        '''
        if self.verbose>1 :     self.show(f"TaskManager - Set running")
        # ---- Get task by id
        task=self.__get_task(task_id)
        # ---- No task, do noything
        if task is None: 
            if self.verbose>1 : self.show(f'    Task is None, do nothing.')
            return
        # ---- Set task as running
        if self.verbose>1 :     self.show(f'    Task set as running ({task_id})')
        task['status'] = 'running'
        task['runner'] = self.whoami
        task['start' ] = now
        task['end'   ] = ''


    # --------------------------------------------------------------
    # -- Set done
    # --------------------------------------------------------------
    #
    def __set_done(self, task_id, now, duration):
        '''
        Set given task_id as "done"
        Complete status and end attributes.
        Params:
            task_id : Task id
        Return:
            None
        '''
        if self.verbose>1 :     self.show(f"TaskManager - Set done")
        # ---- Get task by id
        task=self.__get_task(task_id)
        # ---- No task, do noything
        if task is None: 
            if self.verbose>1 : self.show(f'    Task is None, do nothing.')
            return
        # ---- Set task as done
        if self.verbose>1 :     self.show(f"    Task is set as done ({task_id})")
        task['status']   = 'done'
        task['end'   ]   = now
        task['duration'] = duration
        self.done_tasks += 1


    # --------------------------------------------------------------
    # -- Save task
    # --------------------------------------------------------------
    #
    def __save_tasks(self):
        if self.verbose>1 :     self.show(f'TaskManager - Save tasks')
        with open(self.task_file, 'w', encoding='utf-8') as fp:
            json.dump(self.task_list, fp, ensure_ascii=False, indent=4)
        if self.verbose>1 :     self.show(f'    Saved as : {self.task_file}')



    # --------------------------------------------------------------
    # -- Check dependency
    # --------------------------------------------------------------
    #
    def __check_dependency(self, task):
        # On va chercher une bonne raison pour refuser...

        # Filter is None : No contrains, no reason...
        if task['after'] is None : 
            return True

        # For each entry in the after table 
        for filter in task['after']:

            # We check each task
            for t in self.task_list:
                # Exclude me and desactivated
                if t['id']     == task['id']        : continue
                if t['status'] == 'desactivated'    : continue
                # Doesn't match : next
                if not re.match(filter, t['id'])    : continue
                # Status is not "done" : return False
                if t['status'] != 'done' : return False

        # Arrived here, we found no reason to refuse
        return True




    def update_dependencies(self): 
        '''
        Check and update dependencies. If a task is needed in an after filter, 
        his status will be forced to "ready" 
        Params:
            None
        Returns:
            List of updated tasks
        '''
        if self.verbose>1 :     self.show(f"Update dependancies")
        self.__updated = []
        for task in self.task_list:
            self.__update_dependencies(task)
        return self.__updated



    def __update_dependencies(self, task):
        # Pas active
        if task['status'] != 'ready' : return
        # Pas de dépendances
        if task['after'] is None     : return
        # For each filter in the after list 
        for filter in task['after']:
            # We check each task
            for t in self.task_list:
                # Exclude me
                if task['id'] == t['id']        : continue
                # Doesn't match : next
                if not re.match(filter, t['id']): continue
                # Status have to be ready
                if t['status']!='ready':
                    if self.verbose>0 : self.show(f"Note : For dependency reasons [{t['id']}] is added.")
                    t['status'] = 'ready'
                    self.__updated.append(t['id'])
                # Set dependancies
                self.__update_dependencies(t)




# ------------------------------------------------------------------
#         _______          ____        _ _     _           
#        |__   __|        |  _ \      (_) |   | |          
#            | | ___   ___| |_) |_   _ _| | __| | ___ _ __ 
#            | |/ _ \ / __|  _ <| | | | | |/ _` |/ _ \ '__|
#            | | (_) | (__| |_) | |_| | | | (_| |  __/ |   
#            |_|\___/ \___|____/ \__,_|_|_|\__,_|\___|_|   
#
#                                                   fidle.TocBuilder
# ------------------------------------------------------------------
# A simple class to build a notebook catalog and generate the README.
# Jean-Luc Parouty CNRS/MIAI/SIMaP 2022


from posixpath import dirname
import nbformat
from nbconvert.preprocessors import ExecutePreprocessor
import pandas as pd
from IPython.display import display, Markdown, HTML

import re
import sys, os, glob, yaml
import pathlib
import json
from datetime import datetime
from collections import OrderedDict
from IPython.display import display

import fidle.config as config
from fidle.Chrono import Chrono


class TocBuilder:

    __version__  = config.VERSION



    def __get_files(self, directory, root_dir='.', types=['*.ipynb', '*.py', '*.sh']):
        '''
        Return a list of files to index, from a given directory
        args:
            directory  : dictory to index
            root_dir   : root directory (ie: where directory is)
            types       : Files type to index
        return:
            files : filenames list
        '''
        root_dir  = os.path.expanduser(root_dir)
        all_files = []

        for t in types:
            files = glob.glob(f'{root_dir}/{directory}/{t}')
            files.sort()
            all_files.extend(files)
            
        all_files = [ x.replace(f'{root_dir}/','') for x in all_files]
        return all_files



    def __get_infos(self, filename, root_dir='.'):
        '''
        Extract informations from filename.
        Informations are dirname, basename, id, title, description
        These informations are extracted from comments tags in markdown.
        args:
            filename : file path
            root_dir : root directory 
        return:
            dict : with infos.
        '''
        if filename.endswith('.ipynb'):
            return self.__get_notebook_infos(filename, root_dir)
        else:
            return self.__get_textfile_infos(filename, root_dir)



    def __get_notebook_infos(self, filename, root_dir='.'):
        '''
        Extract informations from a fidle notebook.
        Informations are dirname, basename, id, title, description and are extracted from comments tags in markdown.
        args:
            filename : notebook path from root_dir
            root_dir : root directory 
        return:
            dict : with infos.
        '''
        
        root_dir  = os.path.expanduser(root_dir)

        about={}
        about['id']          = '??'
        about['dirname']     = os.path.dirname(filename)
        about['basename']    = os.path.basename(filename)
        about['title']       = '??'
        about['description'] = '??'
        about['overrides']   = '??'
        
        # ---- Read notebook
        #
        notebook = nbformat.read(f'{root_dir}/{filename}', nbformat.NO_CONVERT)
        
        # ---- Get id, title, desc
        #
        overrides={}
        for cell in notebook.cells:
        
            # ---- Find informations
            #
            if cell['cell_type'] == 'markdown':

                # <!-- TITLE --> tag
                find = re.findall(r'<\!-- TITLE -->\s*\[(.*)\]\s*-\s*(.*)\n',cell.source)
                if find:
                    about['id']    = find[0][0]
                    about['title'] = find[0][1]

                # <!-- DESC --> tag
                find = re.findall(r'<\!-- DESC -->\s*(.*)\n',cell.source)
                if find:
                    about['description']  = find[0]

            # ---- Get override
            #
            if cell['cell_type'] == 'code':
                
                # Try to find : override(...) call
                for m in re.finditer('override\((.+?)\)', cell.source):
                    for override_name in re.findall(r'\w+', m.group(1)):
                        overrides[override_name]='default'

                about['overrides']=overrides

        return about


        
    def __get_textfile_infos(self, filename, root_dir='.'):
        '''
        Extract fidle  informations from a text file (script...).
        Informations are dirname, basename, id, title, description and are extracted from comments tags in document
        args:
            filename : file to analyze
        return:
            dict : with infos.
        '''

        root_dir  = os.path.expanduser(root_dir)

        about={}
        about['id']          = '??'
        about['dirname']     = os.path.dirname(filename)
        about['basename']    = os.path.basename(filename)
        about['title']       = '??'
        about['description'] = '??'
        about['overrides']   = []
        
        # ---- Read file
        #
        with open(f'{root_dir}/{filename}') as fp:
            text = fp.read()

        find = re.findall(r'<\!-- TITLE -->\s*\[(.*)\]\s*-\s*(.*)\n',text)
        if find:
            about['id']    = find[0][0]
            about['title'] = find[0][1]

        find = re.findall(r'<\!-- DESC -->\s*(.*)\n',text)
        if find:
            about['description']  = find[0]

        return about


 

    def __get_content(self, directories={}, root_dir='.'):

        root_dir = os.path.expanduser(root_dir)
        now      = Chrono.now()

        md       = [ f'<!-- Automatically generated on : {now} -->']

        campain= { 'campain' : {  
                        "version"          : '1.0',
                        "description"      : f'Automatically generated ci profile ({now})',
                        "directory"        : './campains/default',
                        "existing_notebook":   "remove    # remove|skip", 
                        "report_template"  :   "fidle     # fidle|default",
                        "timeout"          :   6000
                        }
                 }
        ci = [ yaml.dump(campain, sort_keys=False, default_flow_style=False) ]

        # ------ Notebooks -------------------------------------------------

        # ---- For each directory
        #
        for dir_name,dir_title in directories.items():
            
            # ---- Add title
            md.append(f'\n### {dir_title}')
            ci.append(f'\n#\n# ------------ {dir_name}\n#\n')

            # ---- For each file
            #
            ci_dir = OrderedDict()
            for filename in self.__get_files(dir_name, root_dir):
                
                # Get infos for toc
                #
                infos = self.__get_infos(filename, root_dir)
                file_id          = infos['id']
                file_dirname     = infos['dirname']
                file_basename    = infos['basename']
                file_title       = infos['title']
                file_description = infos['description']
                file_overrides   = infos['overrides']
                file_link        = f'{dir_name}/{file_basename}'.replace(' ','%20')

                md.append(  f'- **[{file_id}]({file_link})** - [{file_title}]({file_link})  '  )
                md.append(  f'{file_description}'  )
            
                # Get infos for ci (if ipynb)
                #
                if not file_basename.endswith('.ipynb'): continue

                ci_file = {}
                ci_file['notebook']  = f'{file_dirname}/{file_basename}'
                if len(file_overrides)>0: ci_file['overrides'] = file_overrides
                
                ci_dir[file_id] = ci_file
            
            # ---- ci / add directory ci
            ci.append( yaml.dump(dict(ci_dir), sort_keys=False, default_flow_style=False) )

        toc_md   = '\n'.join(md)
        toc_ci   = ''.join(ci)
        return toc_md, toc_ci



    def __build_ipynb_readme(self, readme_md='README.md', readme_ipynb='README.ipynb', root_dir='.'):
        '''
        Creation d'un notebook Readme from scratch et sauvegarde de celui-ci
        '''
        # ---- Create Notebook from scratch
        #
        notebook = nbformat.v4.new_notebook()

        # ---- Add a code cell
        #
        code = "from IPython.display import display,Markdown\n"
        code+= "display(Markdown(open('README.md', 'r').read()))\n"
        code+= "#\n"
        code+= "# This README is visible under Jupiter Lab ;-)"
        code+= f"# Automatically generated on : {Chrono.now()}"

        new_cell = nbformat.v4.new_code_cell(source=code)
        new_cell['metadata']= { "jupyter": { "source_hidden": True} }
        notebook.cells.append(new_cell)

        # --- To avoid a modification when opening the notebook
        # # Not annoying in itself, but it requires to save the document again when closing...
        notebook['metadata']["kernelspec"] = {"display_name": "Python 3 (ipykernel)", "language": "python", "name": "python3" }

        # ---- Run it
        #
        ep = ExecutePreprocessor(timeout=600, kernel_name="python3")
        ep.preprocess(notebook,  {'metadata': {'path': root_dir}})

        # ---- Save it
        #
        readme_path = f'{root_dir}/{readme_ipynb}'
        with open(readme_path, mode="w", encoding='utf-8') as fp:
            nbformat.write(notebook, fp)



    def __tag(self,tag, text, document):
        '''
        Put a text inside a tag
        args:
            tag : tag prefix name
            txt : text to insert
            document : document 
        return:
            updated document
        '''
        debut  = f'<!-- {tag}_BEGIN -->'
        fin    = f'<!-- {tag}_END -->'

        output = re.sub(f'{debut}.*{fin}',f'{debut}{text}{fin}',document, flags=re.DOTALL)
        return output




    def update(self, about_file='./fidle/about.yml', root_dir='.'):
        '''
        Update readme with about_file informations
        args:
            about_file : yaml file with fidle informations (./fidle/about.yml)
            root_dir   : root directory (.)
        return:
            Nothing, updated readme is just saved.
        '''

        print(f'\n=== Fidle/TocBuilder - Update tables of content')
        print(f'=== Version : {self.__version__}\n')

        # ---- Preparation
        #
        if about_file is None : 
            print('About_file is None.. exit.')
            return

        root_dir    = os.path.expanduser(root_dir)
        about_file  = os.path.expanduser(about_file)

        # ---- Load about.yml file
        #
        with open(about_file,'r') as fp:
            about=yaml.load(fp, Loader=yaml.FullLoader)

        readme_md    = about['readme_md']
        readme_ipynb = about['readme_ipynb']
        ci_file      = about['default_ci']
        version      = about['version']
        directories  = about['toc']

        print('Update parameters :')
        print('    Root directory   : ', root_dir)
        print('    About file       : ', about_file)

        # ---- Get toc
        #
        print('\nStart update :')
        print(f'    Build toc, for readme and ci...')
        toc_md,toc_ci = self.__get_content(directories, root_dir)

        # ---- Readme markdown -------------------------------------
        #
        # ---- Load readme
        #
        readme_path = f'{root_dir}/{readme_md}'
        print(f'    Load markdown readme...')
        with open(readme_path, 'r') as f:
            readme = f.read()

        # ---- tag version
        #
        print(f'    Update : Version -> {version}')
        readme=self.__tag('VERSION', version, readme)
        
        print(f'    Update : toc')
        readme=self.__tag('TOC', f'\n{toc_md}\n', readme)

        # ---- Save it
        print(f'    Save new markdown readme        : {readme_md}')
        with open(readme_path, 'w') as f:
            f.write(readme)

        # ---- Readme ipynb ----------------------------------------
        #
        print('    Build ipynb readme...')
        self.__build_ipynb_readme(readme_md, readme_ipynb, root_dir)
        print(f'    Save new ipynb readme           : {readme_ipynb}')

        # ---- Save default ci profile -----------------------------
        #
        ci_path = f'{root_dir}/{ci_file}'
        with open(ci_path,'wt') as f:
            f.write(toc_ci)
            print(f'    Save new default ci profile     : {ci_file}')
        print('    done.\n')





# ==================================================================
#                                          _           
#                                         | |          
#              ___  ___ _ __ __ ___      _| | ___ _ __ 
#             / __|/ __| '__/ _` \ \ /\ / / |/ _ \ '__|
#             \__ \ (__| | | (_| |\ V  V /| |  __/ |   
#             |___/\___|_|  \__,_| \_/\_/ |_|\___|_|   
#                                                      
#                                                     fidle.Scrawler
# ==================================================================
# A simple module to host usefull functions for Fidle practical work
# Jean-Luc Parouty CNRS/MIAI/SIMaP 2022


import matplotlib
import matplotlib.pyplot as plt
import math
import itertools
import numpy as np
import sklearn.metrics
import pandas as pd

from IPython.display import display,Image,Markdown,HTML

import fidle.config as config
import fidle.utils  as utils


__version__   = config.VERSION

__save_figs   = False
__figs_dir    = './figs'
__figs_name   = 'fig_'
__figs_id     = 0


# -------------------------------------------------------------
# Plot images
# -------------------------------------------------------------
#
def images(x,y=None, indices='all', columns=12, x_size=1, y_size=1,
                colorbar=False, y_pred=None, cm='binary', norm=None, y_padding=0.35, spines_alpha=1,
                fontsize=20, interpolation='lanczos', save_as='auto'):
    """
    Show some images in a grid, with legends
    args:
        x             : images - Shapes must be (-1,lx,ly) (-1,lx,ly,1) or (-1,lx,ly,3)
        y             : real classes or labels or None (None)
        indices       : indices of images to show or 'all' for all ('all')
        columns       : number of columns (12)
        x_size,y_size : figure size (1), (1)
        colorbar      : show colorbar (False)
        y_pred        : predicted classes (None)
        cm            : Matplotlib color map (binary)
        norm          : Matplotlib imshow normalization (None)
        y_padding     : Padding / rows (0.35)
        spines_alpha  : Spines alpha (1.)
        font_size     : Font size in px (20)
        save_as       : Filename to use if save figs is enable ('auto')
    returns: 
        nothing
    """
    if indices=='all': indices=range(len(x))
    if norm and len(norm) == 2: norm = matplotlib.colors.Normalize(vmin=norm[0], vmax=norm[1])
    draw_labels = (y is not None)
    draw_pred   = (y_pred is not None)
    rows        = math.ceil(len(indices)/columns)
    fig=plt.figure(figsize=(columns*x_size, rows*(y_size+y_padding)))
    n=1
    for i in indices:
        axs=fig.add_subplot(rows, columns, n)
        n+=1
        # ---- Shape is (lx,ly)
        if len(x[i].shape)==2:
            xx=x[i]
        # ---- Shape is (lx,ly,n)
        if len(x[i].shape)==3:
            (lx,ly,lz)=x[i].shape
            if lz==1: 
                xx=x[i].reshape(lx,ly)
            else:
                xx=x[i]
        img=axs.imshow(xx,   cmap = cm, norm=norm, interpolation=interpolation)
#         img=axs.imshow(xx,   cmap = cm, interpolation=interpolation)
        axs.spines['right'].set_visible(True)
        axs.spines['left'].set_visible(True)
        axs.spines['top'].set_visible(True)
        axs.spines['bottom'].set_visible(True)
        axs.spines['right'].set_alpha(spines_alpha)
        axs.spines['left'].set_alpha(spines_alpha)
        axs.spines['top'].set_alpha(spines_alpha)
        axs.spines['bottom'].set_alpha(spines_alpha)
        axs.set_yticks([])
        axs.set_xticks([])
        if draw_labels and not draw_pred:
            axs.set_xlabel(y[i],fontsize=fontsize)
        if draw_labels and draw_pred:
            if y[i]!=y_pred[i]:
                axs.set_xlabel(f'{y_pred[i]} ({y[i]})',fontsize=fontsize)
                axs.xaxis.label.set_color('red')
            else:
                axs.set_xlabel(y[i],fontsize=fontsize)
        if colorbar:
            fig.colorbar(img,orientation="vertical", shrink=0.65)
    save_fig(save_as)
    plt.show()

    
def image(x,cm='binary', figsize=(4,4),interpolation='lanczos', save_as='auto'):
    """
    Draw a single image.
    Image shape can be (lx,ly), (lx,ly,1) or (lx,ly,n)
    args:
        x       : image as np array
        cm      : color map ('binary')
        figsize : fig size (4,4)
    """
    # ---- Shape is (lx,ly)
    if len(x.shape)==2:
        xx=x
    # ---- Shape is (lx,ly,n)
    if len(x.shape)==3:
        (lx,ly,lz)=x.shape
        if lz==1: 
            xx=x.reshape(lx,ly)
        else:
            xx=x
    # ---- Draw it
    plt.figure(figsize=figsize)
    plt.imshow(xx,   cmap = cm, interpolation=interpolation)
    save_fig(save_as)
    plt.show()


# -------------------------------------------------------------
# Plot history
# -------------------------------------------------------------

def history(history, figsize=(8,6), 
                 plot={"Accuracy":['accuracy','val_accuracy'], 'Loss':['loss', 'val_loss']},
                 save_as='auto'):
    """
    Show history
    args:
        history: history
        figsize: fig size
        plot: list of data to plot : {<title>:[<metrics>,...], ...}
    """
    fig_id=0
    for title,curves in plot.items():
        plt.figure(figsize=figsize)
        plt.title(title)
        plt.ylabel(title)
        plt.xlabel('Epoch')
        for c in curves:
            plt.plot(history.history[c])
        plt.legend(curves, loc='upper left')
        if save_as=='auto':
            figname='auto'
        else:
            figname=f'{save_as}_{fig_id}'
            fig_id+=1
        save_fig(figname)
        plt.show()


# -------------------------------------------------------------
# Plot confusion matrix
# -------------------------------------------------------------
def confusion_matrix(y_true,y_pred,
                          target_names,
                          title='Confusion matrix',
                          cmap=None,
                          normalize=True,
                          figsize=(10, 8),
                          digit_format='{:0.2f}',
                          save_as='auto'):
    """
    given a sklearn confusion matrix (cm), make a nice plot

    Arguments
    ---------
    cm:           confusion matrix from sklearn.metrics.confusion_matrix

    target_names: given classification classes such as [0, 1, 2]
                  the class names, for example: ['high', 'medium', 'low']

    title:        the text to display at the top of the matrix

    cmap:         the gradient of the values displayed from matplotlib.pyplot.cm
                  see http://matplotlib.org/examples/color/colormaps_reference.html
                  plt.get_cmap('jet') or plt.cm.Blues

    normalize:    If False, plot the raw numbers
                  If True, plot the proportions

    Citiation
    ---------
    http://scikit-learn.org/stable/auto_examples/model_selection/plot_confusion_matrix.html

    """
    cm = sklearn.metrics.confusion_matrix( y_true,y_pred, normalize=None, labels=target_names)
    
    accuracy = np.trace(cm) / float(np.sum(cm))
    misclass = 1 - accuracy

    if cmap is None:
        cmap = plt.get_cmap('Blues')

    plt.figure(figsize=figsize)
    plt.imshow(cm, interpolation='nearest', cmap=cmap)
    plt.title(title)
    plt.colorbar()

    if target_names is not None:
        tick_marks = np.arange(len(target_names))
        plt.xticks(tick_marks, target_names, rotation=90)
        plt.yticks(tick_marks, target_names)

    if normalize:
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]


    thresh = cm.max() / 1.5 if normalize else cm.max() / 2
    for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
        if normalize:
            plt.text(j, i, digit_format.format(cm[i, j]),
                     horizontalalignment="center",
                     color="white" if cm[i, j] > thresh else "black")
        else:
            plt.text(j, i, "{:,}".format(cm[i, j]),
                     horizontalalignment="center",
                     color="white" if cm[i, j] > thresh else "black")

    plt.tight_layout()
    plt.ylabel('True label')
    plt.xlabel('Predicted label\naccuracy={:0.4f}; misclass={:0.4f}'.format(accuracy, misclass))
    save_fig(save_as)
    plt.show()


    
def confusion_matrix_txt(y_true,y_pred,labels=None,color='green',
                             font_size='12pt', title="#### Confusion matrix is :"):
    """
    Show a confusion matrix for a predictions.
    see : sklearn.metrics.confusion_matrix

    Args:
        y_true:       Real classes
        y_pred:       Predicted classes
        labels:       List of classes to show in the cm
        color:        Color for the palette (green)
        font_size:    Values font size 
        title:        the text to display at the top of the matrix        
    """
    assert (labels!=None),"Label must be set"
    
    if title != None :  display(Markdown(title)) 
    
    cm = sklearn.metrics.confusion_matrix( y_true,y_pred, normalize="true", labels=labels)
    df=pd.DataFrame(cm)

#     cmap = sn.light_palette(color, as_cmap=True)

    colorsList = ['whitesmoke','bisque']
    cmap = matplotlib.colors.ListedColormap(colorsList)
    cmap = matplotlib.colors.ListedColormap(cmap(np.linspace(0, 1, 256)))

    df.style.set_properties(**{'font-size': '20pt'})
    display(df.style.format('{:.2f}') \
            .background_gradient(cmap=cmap)
            .set_properties(**{'font-size': font_size}))
    

# -------------------------------------------------------------
# Plot donut
# -------------------------------------------------------------

def donut(values, labels, colors=["lightsteelblue","coral"], figsize=(6,6), title=None, save_as='auto'):
    """
    Draw a donut
    args:
        values   : list of values
        labels   : list of labels
        colors   : list of color (["lightsteelblue","coral"])
        figsize  : size of figure ( (6,6) )
    return:
        nothing
    """
    # ---- Title or not
    if title != None :  display(Markdown(title))
    # ---- Donut
    plt.figure(figsize=figsize)
    # ---- Draw a pie  chart..
    plt.pie(values, labels=labels, 
            colors = colors, autopct='%1.1f%%', startangle=70, pctdistance=0.85,
            textprops={'fontsize': 18},
            wedgeprops={"edgecolor":"w",'linewidth': 5, 'linestyle': 'solid', 'antialiased': True})
    # ---- ..with a white circle
    circle = plt.Circle((0,0),0.70,fc='white')
    ax = plt.gca()
    ax.add_artist(circle)
    # Equal aspect ratio ensures that pie is drawn as a circle
    plt.axis('equal')  
    plt.tight_layout()
    save_fig(save_as)
    plt.show()
    

    
def multivariate_serie(sequence, labels=None, predictions=None, only_features=None,
                            columns=3, width=5,height=4,wspace=0.3,hspace=0.2,ms=6,lw=1,
                            save_as='auto', time_dt=1, hide_ticks=False):
    
    sequence_len = len(sequence)
    features_len = sequence.shape[1]
    if only_features is None : only_features=range(features_len)
    if labels is None        : labels=range(features_len)
    
    t  = np.arange(sequence_len)    
    if predictions is None:
        dt = 0
    else:
        dt = len(predictions)
        sequence_with_pred = sequence.copy()
        sequence_with_pred[-dt:]=predictions

    rows = math.ceil(features_len/columns)
    fig  = plt.figure(figsize=(columns*width, rows*height))
    fig.subplots_adjust(wspace=0.3,hspace=0.2)
    n=1
    for i in only_features:
        ax=fig.add_subplot(rows, columns, n)
        
        # ---- Real sequence without prediction
        #
        ax.plot( t[:-dt],sequence[:-dt,i], 'o',  markersize=ms, color='C0', zorder=2)
        ax.plot( t,sequence[:,i],          '-',  linewidth=lw,  color='C0', label=labels[i],zorder=1)

        # ---- What we expect
        #
        ax.plot(t[-dt:], sequence[-dt:,i], 'o', markeredgecolor='C0',markerfacecolor='white',ms=6)
        
        if predictions is not None:
            ax.plot(t[-dt-1:], sequence_with_pred[-dt-1:,i], '--',  lw=lw, fillstyle='full',  ms=ms, color='C1',zorder=1)
            ax.plot(t[-dt:],   predictions[:,i],             'o',   lw=lw, fillstyle='full',  ms=ms, color='C1',zorder=2)

        if hide_ticks:
            ax.set_yticks([])
            ax.set_xticks([])
        
        ax.legend(loc="upper left")
        n+=1
    save_fig(save_as)
    plt.show()

    
# -------------------------------------------------------------
# Show 2d series and segments
# -------------------------------------------------------------
#

def serie_2d(data, figsize=(10,8), monocolor=False, hide_ticks=True, lw=2, ms=4, save_as='auto'):
    """
    Plot a 2d dataset as a trajectory
    args:
        data:      Dataset to plot
        figsize:   Figure size ( (10,8))
        monocolor: Monocolor line or not. (False)
    return:
        nothing
    """
    # ---- Get x,y, min and max
    #
    n     = len(data)
    k     = int(n/100)
    x,y   = data[:,0], data[:,1]
 
    # ---- Draw it
    #
    fig = plt.figure(figsize=figsize)
    ax = plt.axes()

    # ---- Monocolor or gray gradient
    #
    if monocolor:
        ax.plot(x,y)
    else:
        for i in range(0,100):
            a= (200-i)/200
            ax.plot(x[i*k:(i+1)*k+1], y[i*k:(i+1)*k+1], '-', color=(a,a,a),lw=lw,zorder=1)

    # ---- Last point
    #
    ax.plot(x[n-1], y[n-1], 'o', color='C1',ms=ms,zorder=2)
    
    ax.set_aspect('equal', 'box')
    ax.set_xlabel('axis=0')
    ax.set_ylabel('axis=1')
    
    if hide_ticks:
        ax.set_yticks([])
        ax.set_xticks([])

    save_fig(save_as)
    plt.show()
    
    
def segment_2d(sequence_real, sequence_pred, figsize=(10,8), ms=6, lw=1, hide_ticks=True, save_as='auto'):
    """
    Plot a 2d segment real and predicted
    args:
        sequence_real: Real sequence
        sequence_pred: Predicted sequence
        figsize:       Figure size ( (10,8) )
        ms:            Marker size (6)
    return:
        nothing
    """
    k = len(sequence_pred)
    x,y = sequence_real[:,0],sequence_real[:,1]
    u,v = sequence_pred[:,0],sequence_pred[:,1]
    
    fig = plt.figure(figsize=figsize)

    ax = plt.axes()
    
    # ---- Draw real sequence without prediction
    #
    ax.plot(x[:-k], y[:-k],   'o', color='C0', fillstyle='full', zorder=2, ms=ms)
    ax.plot(x, y,             '-', color='C0', lw=lw, zorder=1)
    
    # ---- What we expect
    #
    ax.plot(x[-k:], y[-k:], 'o', ms=ms, markeredgecolor='C0', markerfacecolor='white', zorder=2)

    # ---- What we have
    #
    ax.plot(u, v,                            'o',  color='C1',fillstyle='full',zorder=2, ms=ms)
    ax.plot( [x[-1-k],u[0]], [y[-1-k],v[0]], '--', color='C1',lw=lw, zorder=1)
    ax.plot(u, v,                            '--', color='C1',lw=lw, zorder=1)

    ax.set_aspect('equal', 'box')
    ax.set_xlabel('axis=0')
    ax.set_ylabel('axis=1')
    
    if hide_ticks:
        ax.set_yticks([])
        ax.set_xticks([])

    save_fig(save_as)
    plt.show()

    
    

# -------------------------------------------------------------
# Saving figs stuff
# -------------------------------------------------------------
   
def set_save_fig(save=True, figs_dir='./figs', figs_name='fig_', figs_id=0):
    """
    Set save_fig parameters
    Default figs name is <figs_name><figs_id>.{png|svg}
    args:
        save      : Boolean, True to save figs (True)
        figs_dir  : Path to save figs (./figs)
        figs_name : Default basename for figs (figs_)
        figs_id   : Start id for figs name (0)
    """
    global __save_figs, __figs_dir, __figs_name, __figs_id
    __save_figs = save
    __figs_dir  = figs_dir
    __figs_name = figs_name
    __figs_id   = figs_id
    
    
def save_fig(filename='auto', png=True, svg=False):
    """
    Save current figure
    args:
        filename : Image filename ('auto')
        png      : Boolean. Save as png if True (True)
        svg      : Boolean. Save as svg if True (False)
    """
    global __figs_id

    if filename is None  : return
    if not __save_figs   : return

    utils.mkdir(__figs_dir)
    
    if filename=='auto': 
        path=f'{__figs_dir}/{__figs_name}{__figs_id:02d}'
        __figs_id+=1
    else:
        path=f'{__figs_dir}/{filename}'

    if png : plt.savefig( f'{path}.png')
    if svg : plt.savefig( f'{path}.png')
    
    utils.display_html(f'<div class="comment">Saved: {path}</div>')
    




# ------------------------------------------------------------------
#   ______ _     _ _        _____           _        _ _           
#  |  ____(_)   | | |      |_   _|         | |      | | |          
#  | |__   _  __| | | ___    | |  _ __  ___| |_ __ _| | | ___ _ __ 
#  |  __| | |/ _` | |/ _ \   | | | '_ \/ __| __/ _` | | |/ _ \ '__|
#  | |    | | (_| | |  __/  _| |_| | | \__ \ || (_| | | |  __/ |   
#  |_|    |_|\__,_|_|\___| |_____|_| |_|___/\__\__,_|_|_|\___|_|   
#
#                                                    Fidle installer
# ------------------------------------------------------------------
# A simple class to install fidle notebooks and datasets
# Jean-Luc Parouty CNRS/MIAI/SIMaP 2022


import os, sys
import importlib
import yaml
import requests, json
import tarfile
from random import randint

from IPython.display import display

import fidle.config       as     config
import fidle.utils        as     utils


class Installer:

    __version__  = config.VERSION


    def __init__(self):
        
        try:
            self.catalog = self.__load_catalog()
        except:
            print('\n*** OUPS ! The list of available installations cannot be retrieved...')
            print("*** Maybe it's a network problem or a Fidle problem.\n")



    def __load_catalog(self):
        '''
        Load the catalog od ressources
        args:
            None
        return:
            The catalog as a dict
        '''
        url = requests.get(config.CATALOG_URL)
        data = json.loads(url.text)
        return data


    def __check_catalog_name(self, catalog_name):
        '''
        Check catalog name
        args:
            catalog_name : Catalalog name
        return:
            True if catalog exist else False
        '''
        if not f'{catalog_name}.list' in self.catalog:
            print(f'\n*** Oups, The requested catalog "{catalog_name}" cannot be found...')
            return False
        return True


    def __check_directory(self, directory='.'):
        '''
        Check if directory exist and is writable
        args:
            directory :  directory to check
        return:
            True if installation directory exist else False
        '''
        if not os.path.isdir(directory):
            print('\n*** Oups, the installation directory cannot be found...\n')
            return False
        if not os.access(directory, os.W_OK):
            print('\n*** Oups, the installation directory is not writeable...\n')
            return False
        return True


    def __download_ressource(self, url, installation_dir='.', verbose=1):
        '''
        Download a ressource under a random prefixed name.
        args:
            url : Ressource URL
            installation_dir : The place to download the resource (.)
        return:
            Path to the downloaded ressource
        '''
        # ---- Check intallation_dir
        #
        installation_dir = os.path.expanduser(installation_dir)
        if not self.__check_directory(installation_dir): return False

        # ---- Retrieve all infos
        #
        ressource_name   = os.path.basename(url)
        prefix           = str(randint(1000000, 9999999))
        ressource_path   = f'{installation_dir}/_{prefix}_{ressource_name}'

        # ---- Download ressource
        #
        with open(ressource_path, "wb") as fp:
            response     = requests.get(url, stream=True)
            total_length = response.headers.get('content-length')

            if total_length is None: # no content length header
                fp.write(response.content)
            else:
                total_length = int(total_length)
                total_hsize  = utils.hsize(total_length)
                dl   = 0
                i    = 0
                imax = int(total_length/4096)
                for data in response.iter_content(chunk_size=4096):
                    fp.write(data)
                    if verbose>0: utils.update_progress('Download : ',i,imax, redraw=False, verbosity=2, total_max=total_hsize)
                    i+=1
        return ressource_path


    def __extract_ressource(self, ressource_path, installation_dir='.', verbose=1):
        '''
        Extract a downloded ressource to an installation_dir
        args:
            ressource path : Path to the ressource tar archive
            installation_dir : The place to extract 
        return:
            installation dir
        '''
        # ---- Check intallation_dir
        #
        installation_dir = os.path.expanduser(installation_dir)
        if not self.__check_directory(installation_dir): return False

        # ---- Tar extraction
        #
        with tarfile.open(name=ressource_path) as tar:

            imax = len(tar.getmembers())
            i    = 0
            total_files = f'{imax} files'
            for member in tar.getmembers():
                i+=1
                tar.extract(member=member, path=installation_dir)
                if verbose>0: utils.update_progress('Extract  :',i,imax, redraw=False, verbosity=2, total_max=total_files)
        
        return installation_dir




    def show_branch(self,catalog_name=None):
        '''
        Print a description of available resources.
        args:
            catalog_name : Catalog name,  'notebooks'|'datasets'
        return:
            A textual description on the catalog or False
        '''
        # ---- Get catalog
        #
        if not self.__check_catalog_name(catalog_name): return False

        catalog_list    = self.catalog[ f'{catalog_name}.list']
        catalog_default = self.catalog[ f'{catalog_name}.default']

        # ---- Build description
        #
        about = f'Installable resources in catalog [{catalog_name}]\n\n'
        for entry in catalog_list:
            name = entry['name']
            desc = entry['description']
            if name == catalog_default : name = name+'*'
            about += f'   - {name:20s} :  {desc}\n'

        # ---- Print or display
        #
        print(about)



    def install_ressource(self, catalog_name=None, ressource_name=None, installation_dir='.', verbose=1):
        '''
        Install a ressource
        args:
            catalog_name     : Catalog name, 'notebooks'|'datasets'
            ressource_name   : Resource name in the catalog
            installation_dir : The location where to install the requested resource
        return:
            True if the install is ok, else False
        '''

        # ---- Welcome
        #
        if verbose>1:
            print(f'\n=== Fidle/Installer - Install ')
            print(f'=== Version : {self.__version__}')

        # ---- Check catalog_name
        #
        if verbose>1: print('Check requested ressource...')
        if not self.__check_catalog_name(catalog_name): return False

        # ---- Default ressource ?
        #
        if ressource_name=='default':
            ressource_name = self.catalog[ f'{catalog_name}.default' ]
        
        # ---- Check intallation_dir
        #
        if verbose>1: print('Check installation directory...')
        installation_dir = os.path.expanduser(installation_dir)
        if not self.__check_directory(installation_dir): return False

        if os.path.isdir( f'{installation_dir}/{ressource_name}' ):
            print(f'\n*** Oups, the installation directory already contains a directory "{ressource_name}"...\n')
            return

        # ---- Try to find the ressource url/name
        #
        found=False
        for entry in self.catalog[ f'{catalog_name}.list']:
            if entry['name'] == ressource_name:
                found=True
                break
        
        if not found:
            print('\n*** Oups, the requested resource is not in the catalog...\n')
            return

        ressource_url  = entry['url']

        # ---- Download and install ressource
        #
        if verbose>0: 
            print(f'Install ressource : {ressource_name}')
            print(f'In directory      : {installation_dir}')
        ressource_path   = self.__download_ressource(ressource_url, installation_dir, verbose=verbose)
        installation_dir = self.__extract_ressource(ressource_path, installation_dir, verbose=verbose)

        os.remove(ressource_path)
        print('Done.\n')
        # 


    def check(self, directory='.'):
        
        # ---- Clean/expand directory
        #
        directory = os.path.expanduser(directory)
        directory = os.path.abspath(directory)

        # ---- Search notebooks and datasets
        # le dossier exploré
        print('Notebooks and datasets can only be found if they are in/near the explored folder.')
        print(f'Explored directory is : {directory}')

        # datasets_dir by env var
        datasets_dir = os.getenv('FIDLE_DATASETS_DIR', 'undefined')
        # Searching
        notebooks, datasets = utils.looking_about(path=directory)

        # ---- Check datasets
        # 
        print('\nDatasets dir found : \n')

        for d in datasets:
            print(f"    {d['path']:50s}   ({d['name']} / {d['version']})")

        print(f'    The environment variable FIDLE_DATASETS_DIR is :     {datasets_dir}')

        if len(datasets)==0 and datasets_dir=='undefined':
            print('\n    ** No datasets directory found...')
            print('    ** Have you installed the notebooks ?')
            print('    ** Are you in / do you check the right directory ?')

        # ---- Check notebooks
        #
        print('\nNotebooks dir found : \n')


        for d in notebooks:
            print(f"    {d['path']:50s}   ({d['name']} / {d['version']})")

        if len(notebooks)==0:
            print('    ** No notebooks directory found...')
            print('    ** Have you installed the notebooks?')
            print('    ** Are you in /do ou check the right directory ?')

        # ---- Check environment
        #
        print('\nCheck environment : \n')
        print(f'    Fidle module         : Ok         ({config.VERSION})')

        for mod in config.USED_MODULES:
            try:
                importlib.import_module(mod)
                print(f'    {mod:21s}: Ok         ({sys.modules[mod].__version__})')
            except:
                print(f'    {mod:21s}: Not found')
        print('')
